import os
import collections
import bisect
import io
import json

import numpy as np

from . import errors
from . import utils
from . import style_bits
from functools import reduce

import logging
log = logging.getLogger(__name__)


class ArrayWrapper:
    """Wrapper for numpy data so that manipulations can use normal numpy syntax
    and still affect the data according to the byte ordering.

    Numpy's fancy indexing can't be used for setting set values, so this
    intermediate layer is needed that defines the __setitem__ method that
    explicitly references the byte ordering in the data array.
    """

    def __init__(self, data, order):
        self.np_data = data
        self.order = order

    def __str__(self):
        return f"ArrayWrapper at {hex(id(self))} count={len(self)} order={self.order}"

    def __len__(self):
        return np.alen(self.order)

    def __and__(self, other):
        return self.np_data[self.order] & other

    def __iand__(self, other):
        self.np_data[self.order] &= other
        return self

    def __or__(self, other):
        return self.np_data[self.order] | other

    def __ior__(self, other):
        self.np_data[self.order] |= other
        return self

    def __getitem__(self, index):
        return self.np_data[self.order[index]]

    def __setitem__(self, index, value):
        self.np_data[self.order[index]] = value

    def tobytes(self):
        return self.np_data[self.order].tobytes()


class Segment:
    ui_name = "Data Segment"
    base_serializable_attributes = ['origin', 'error', 'name', 'verbose_name', 'uuid']
    extra_serializable_attributes = []
    dependent_file_attributes = []

    def __init__(self, container_or_segment, offset_or_offset_list=None, origin=0, name="All", error=None, verbose_name=None, length=None):
        self.init_empty()
        # the container may be specified as the actual container or a segment
        # of the container. If a segment is specified, the offset list is
        # calculated relative to the segment to get the real offset into the
        # container.
        offset_list = self.calc_offset_list(container_or_segment, offset_or_offset_list, length)
        if hasattr(container_or_segment, 'container_offset'):
            log.debug(f"creating {name},  {len(offset_list)} bytes from {container_or_segment}")
            # log.debug(f"  offset_list = {offset_list}")
            offset_list = container_or_segment.container_offset[offset_list]
            container_or_segment = container_or_segment.container

        self.container = container_or_segment
        self.container_offset = self.enforce_offset_bounds(offset_list)

        self.origin = int(origin)  # force python int to decouple from possibly being a numpy datatype
        self.error = error
        self.name = name
        self.verbose_name = verbose_name

        # Child segments
        self.segments = self.calc_segments()

    def init_empty(self):
        self.origin = 0
        self.error = ""
        self.name = ""
        self.verbose_name = ""
        self.uuid = utils.uuid()
        self._reverse_offset = None
        self.segments = []

    #### properties

    @property
    def data(self):
        return ArrayWrapper(self.container._data, self.container_offset)

    @property
    def style(self):
        return ArrayWrapper(self.container._style, self.container_offset)

    @property
    def disasm_type(self):
        return ArrayWrapper(self.container._disasm_type, self.container_offset)

    @property
    def reverse_offset(self):
        if self._reverse_offset is None:
            self._reverse_offset = self.calc_reverse_offsets()
        return self._reverse_offset

    def __len__(self):
        return np.alen(self.container_offset)

    #### dunder methods and convenience functions to operate on data (not style)

    def __str__(self):
        if self.origin > 0:
            origin = " @ %04x" % (self.origin)
        else:
            origin = ""
        s = "%s (%d bytes%s)" % (self.name, len(self), origin)
        if self.error:
            s += " " + self.error
        return s

    def __and__(self, other):
        return self.container._data[self.container_offset] & other

    def __iand__(self, other):
        self.container._data[self.container_offset] &= other
        return self

    def __getitem__(self, index):
        return self.container._data[self.container_offset[index]]

    def __setitem__(self, index, value):
        self.container._data[self.container_offset[index]] = value

    #### iterator utilities

    def iter_segments(self, segment_type=None):
        for segment in self.segments:
            if segment_type is None or isinstance(segment, segment_type):
                yield segment
            yield from segment.iter_segments(segment_type)

    def iter_menu(self, level):
        for segment in self.segments:
            yield (segment, level)
            yield from segment.iter_menu(level + 1)

    #### dependent files

    @property
    def has_dependent_files(self):
        return bool(self.dependent_file_attributes)

    @property
    def dependent_files(self):
        """List of all dependent files attached to the segment.

        Each attribute in the dependent_file_attributes list can either return
        a filename or a list of filenames, so this property gathers all of
        those into a single list.
        """
        files = []
        for key in self.dependent_file_attributes:
            value = getattr(self, key)
            if value is None:
                continue
            elif isinstance(value, str):
                files.append(value)
            else:
                files.extend(value)
        if not hasattr(self, 'dependent_file_timestamps'):
            self.dependent_file_timestamps = collections.defaultdict(int)
        return files

    def check_dependent_files(self):
        """Check timestamps of dependent files and trigger an update if the
        file has changed since the last check.

        Note that there will always be at least one update event the first time
        a dependent file is seen, because the update timestamp is set to zero
        (unix time corresponding to Jan 1, 1970) on initialization.
        """
        log.debug(f"check_dependent_files: segment={self}")
        for path in self.dependent_files:
            try:
                current = os.stat(path).st_mtime
            except IOError as e:
                self.error_in_dependent_file(path, e)
            else:
                previous = self.dependent_file_timestamps[path]
                if current > previous:
                    log.debug(f"check_dependent_files: {path}, old={previous}, new={current}")
                    self.update_dependent_file(path)
                    self.dependent_file_timestamps[path] = current

    def update_dependent_file(self, path):
        """Used in subclasses to perform an action on dependent file.

        path: the pathname of the changed file
        """
        pass

    def error_in_dependent_file(self, path, error):
        """Used in subclasses to report and error in a particular file from the
        list of dependent files.

        path: the pathname of the changed file
        error: the error as reported by python
        """
        pass

    #### offsets

    def calc_offset_list(self, container_or_segment, offset_or_offset_list=None, length=None):
        if offset_or_offset_list is None:
            offset_or_offset_list = 0
        try:
            start_offset = int(offset_or_offset_list)
        except TypeError:
            offset_list = utils.to_numpy_list(offset_or_offset_list)
        else:
            if length is None:
                length = len(container_or_segment)
            offset_list = np.arange(offset_or_offset_list, offset_or_offset_list + length, dtype=np.uint32)
        return offset_list

    def enforce_offset_bounds(self, offset_list):
        return offset_list[offset_list < len(self.container)]

    def calc_reverse_offsets(self):
        # Initialize array to out of range
        r = np.zeros(len(self.container), dtype=np.int32) - 1
        r[self.container_offset] = np.arange(len(self), dtype=np.int32)
        valid = np.where(r >= 0)[0]
        if len(valid) != len(self):
            raise errors.InvalidSegmentOrder
        return r

    def calc_offsets_of_range(self, start, end):
        in_here = np.arange(start, end, dtype=np.int32)
        in_container = self.container_offset[in_here]
        return in_container

    def calc_index_from_other_segment(self, other_segment_index, other_segment):
        """Convert an index from another segment by mapping it to the container
        and then looking up the index that corresponds to that container index
        in this segment.
        """
        container_index = other_segment.container_offset[other_segment_index]
        index = self.reverse_offset[container_index]
        return index

    def calc_indexes_from_other_segment(self, other_segment_indexes, other_segment):
        """Convert a list of indexes from another segment by mapping it to the
        container and then looking up each index that corresponds to that
        container index in this segment.
        """
        container_indexes = other_segment.container_offset[other_segment_indexes]
        indexes = self.reverse_offset[container_indexes]
        return indexes

    #### creation

    def calc_segments(self):
        """Convenience method used by subclasses to create any sub-segments
        within this segment.

        """
        return []

    #### serialization

    def __getstate__(self):
        """Custom jsonpickle state save routine

        This routine culls down the list of attributes that should be
        serialized, and in some cases changes their format slightly so they
        have a better mapping to json objects. For instance, json can't handle
        dicts with integer keys, so dicts are turned into lists of lists.
        Tuples are also turned into lists because tuples don't have a direct
        representation in json, while lists have a compact representation in
        json.
        """
        state = dict()
        state['__size__'] = len(self)

        def get_value(key):
            if ":" in key:
                key, converter = key.split(":", 1)
                value = getattr(self, key)
                if converter == "int":
                    value = int(value)
                elif converter == "str":
                    value = str(value)
            else:
                value = getattr(self, key)
            return key, value

        for key in self.base_serializable_attributes + self.extra_serializable_attributes + self.dependent_file_attributes:
            key, value = get_value(key)
            state[key] = value
        state['container_offset'] = utils.collapse_to_ranges(self.container_offset, compact=True)
        state['segments'] = self.segments
        return state

    def __setstate__(self, state):
        """Custom jsonpickle state restore routine

        The use of jsonpickle to recreate objects doesn't go through __init__,
        so there will be missing attributes when restoring old versions of the
        json. Once a version gets out in the wild and additional attributes are
        added to a segment, a default value should be applied here.
        """
        self.init_empty()
        size = state.pop('__size__')
        raw = np.arange(size, dtype=np.uint32)
        self.container_offset = raw
        utils.restore_from_ranges(self.container_offset, state.pop('container_offset', []))
        self.segments = state.pop('segments')

        # Can't restore here because it would result in many unrelated copies
        # of the container object. After all the segments are loaded, will have
        # to loop through and set container attributes for each one.
        self.container = None

        for key in self.base_serializable_attributes + self.extra_serializable_attributes + self.dependent_file_attributes:
            if key in state:
                setattr(self, key, state.pop(key))
        self.restore_missing_serializable_defaults()
        self.__dict__.update(state)
        self.restore_renamed_serializable_attributes()

    def restore_missing_serializable_defaults(self):
        """Hook for the future when extra serializable attributes are added to
        subclasses so new versions of the code can restore old saved files by
        providing defaults to any missing attributes.
        """
        pass

    def restore_renamed_serializable_attributes(self):
        """Hook for the future if attributes have been renamed. The old
        attribute names will have been restored in the __dict__.update in
        __setstate__, so this routine should move attribute values to their new
        names.
        """
        if hasattr(self, 'start_addr'):
            self.origin = self.start_addr
            log.debug(f"moving start_addr to origin: {self.start_addr}")
            delattr(self, 'start_addr')

    def restore_computed_defaults(self):
        """Hook to recreate any computed defaults after the container has
        been restored
        """
        pass


    @property
    def verbose_info(self):
        lines = []
        name = self.verbose_name or self.name
        lines.append(f"{name}: {len(self)} bytes")
        for s in self.segments:
            v = s.segment_info("    ")
            lines.extend(v)
        return "\n".join(lines)

    def segment_info(self, indent=""):
        lines = []
        lines.append(indent + str(self))
        for s in self.segments:
            lines.extend(s.segment_info(indent + "    "))
        return lines

    def is_valid_index(self, i):
        return i >= 0 and i < len(self)

    def tobytes(self):
        return self.container._data[self.container_offset].tobytes()

    def calc_source_indexes_from_ranges(self, ranges):
        source_indexes = np.zeros(len(self.container), dtype=np.uint8)
        offsets = self.container_offset
        for start, end in ranges:
            if end < start:
                start, end = end, start
            source_indexes[offsets[start:end]] = 1
        affected_source_indexes = np.where(source_indexes > 0)[0]
        return affected_source_indexes

    def set_style_ranges(self, ranges, **kwargs):
        indexes = self.calc_source_indexes_from_ranges(ranges)
        self.container.set_style_at_indexes(indexes, **kwargs)

    def clear_style_ranges(self, ranges, **kwargs):
        indexes = self.calc_source_indexes_from_ranges(ranges)
        self.container.clear_style_at_indexes(indexes, **kwargs)

    def clear_style_bits(self, **kwargs):
        self.container.clear_style_at_indexes(self.container_offset, **kwargs)

    def get_style_ranges(self, **kwargs):
        """Return a list of start, end pairs that match the specified style
        """
        bits = style_bits.get_style_bits(**kwargs)
        matches = (self.style & bits) == bits
        return utils.bool_to_ranges(matches)

    def get_style_indexes(self, **kwargs):
        """Return a list of start, end pairs that match the specified style
        """
        bits = style_bits.get_style_bits(**kwargs)
        matches = (self.style & bits) == bits
        w = np.where(matches == True)[0]
        return w

    def convert_style(self, from_style, to_style):
        indexes = self.get_style_indexes(**from_style)
        indexes = self.container_offset[indexes]
        c = self.container
        c.clear_style_at_indexes(indexes, **from_style)

        # it's possible the mask includes more bits than the style, as in the
        # user styles
        c.clear_style_at_indexes(indexes, **to_style)
        c.set_style_at_indexes(indexes, **to_style)

    #### disassembly type

    def update_data_style_from_disasm_type(self):
        self.container.update_data_style_from_disasm_type()

    def set_disasm_ranges(self, ranges, value):
        indexes = self.calc_source_indexes_from_ranges(ranges)
        self.container.disasm_type[indexes] = value

    #### comment convenience functions

    def get_comment_locations(self, **kwargs):
        bits = style_bits.get_style_bits(**kwargs)
        r = self.rawdata.copy()
        #print len(r.style)
        #print len(r.style_base)
        r.style_base[:] &= bits
        comment_indexes = np.asarray(list(self.container.comments.keys()), dtype=np.uint32)
        #print comment_indexes
        r.style_base[comment_indexes] |= style_bits.comment_bit_mask
        return r.unindexed_style[:]

    def find_next(self, index, **kwargs):
        ranges = self.get_style_ranges(**kwargs)
        if len(ranges) > 0:
            index_tuple = (index + 1, 0)
            match_index = bisect.bisect_right(ranges, index_tuple)
            if match_index >= len(ranges):
                match_index = 0
            return ranges[match_index][0]
        return None

    def find_previous(self, index, **kwargs):
        ranges = self.get_style_ranges(**kwargs)
        if len(ranges) > 0:
            index_tuple = (index - 1, 0)
            match_index = bisect.bisect_left(ranges, index_tuple)
            match_index -= 1
            if match_index < 0:
                match_index = len(ranges) - 1
            return ranges[match_index][0]
        return None

    def remove_comments_at_indexes(self, indexes):
        for where_index in indexes:
            self.remove_comment(where_index)

    def set_comments_at_indexes(self, ranges, indexes, comments):
        c = self.container
        for where_index, comment in zip(indexes, comments):
            rawindex = self.container_offset[where_index]
            if comment:
                log.debug("  restoring comment: rawindex=%d, '%s'" % (rawindex, comment))
                c.comments[rawindex] = comment
            else:
                try:
                    del self.container.comments[rawindex]
                    log.debug("  no comment in original data, removed comment in current data at rawindex=%d" % rawindex)
                except KeyError:
                    log.debug("  no comment in original data or current data at rawindex=%d" % rawindex)
                    pass
        c.fixup_comments()

    def get_comments_at_indexes(self, indexes):
        """Get a list of comments at specified indexes"""
        s = self.style[indexes]
        has_comments = np.where(s & style_bits.comment_bit_mask > 0)[0]
        comments = []
        for where_index in has_comments:
            raw = self.container_offset[indexes[where_index]]
            try:
                comment = self.container.comments[raw]
            except KeyError:
                comment = None
            comments.append(comment)
        return has_comments, comments

    def get_comment_restore_data(self, ranges):
        """Get a chunk of data (designed to be opaque) containing comments,
        styles & locations that can be used to recreate the comments on an undo
        """
        restore_data = []
        for start, end in ranges:
            log.debug("range: %d-%d" % (start, end))
            styles = self.style[start:end].copy()
            items = {}
            for i in range(start, end):
                rawindex = self.container_offset[i]
                try:
                    comment = self.container.comments[rawindex]
                    log.debug("  index: %d rawindex=%d '%s'" % (i, rawindex, comment))
                    items[i] = (rawindex, comment)
                except KeyError:
                    log.debug("  index: %d rawindex=%d NO COMMENT TO SAVE" % (i, rawindex))
                    items[i] = (rawindex, None)

            restore_data.append((start, end, styles, items))
        return restore_data

    def restore_comments(self, restore_data):
        """Restore comment styles and data
        """
        for start, end, styles, items in restore_data:
            log.debug("range: %d-%d" % (start, end))
            self.style[start:end] = styles
            for i in range(start, end):
                rawindex, comment = items[i]
                if comment:
                    log.debug("  restoring comment: rawindex=%d, '%s'" % (rawindex, comment))
                    self.container.comments[rawindex] = comment
                else:
                    # no comment in original data, remove any if exists
                    try:
                        del self.container.comments[rawindex]
                        log.debug("  no comment in original data, removed comment in current data at rawindex=%d" % rawindex)
                    except KeyError:
                        log.debug("  no comment in original data or current data at rawindex=%d" % rawindex)
                        pass

    def get_comments_in_range(self, start, end):
        """Get a list of comments at specified indexes"""
        comments = {}

        # Naive way, but maybe it's fast enough: loop over all comments
        # gathering those within the bounds
        for rawindex, comment in self.container.comments.items():
            try:
                index = self.get_index_from_base_index(rawindex)
            except IndexError:
                continue
            if index >= start and index < end:
                comments[index] = comment
        return comments

    def set_comment_at(self, index, text):
        rawindex = self.container_offset[index]
        c = self.container
        c.comments[rawindex] = text
        c.style[rawindex] |= style_bits.comment_bit_mask

    def set_comment_ranges(self, ranges, text):
        for start, end in ranges:
            self.set_comment_at(start, text)

    def get_comment_at(self, index):
        rawindex = self.container_offset[index]
        return self.container.comments.get(rawindex, "")

    def remove_comment_at(self, index):
        rawindex = self.container_offset[index]
        self.container.clear_comments([rawindex])

    def get_first_comment(self, ranges):
        start = reduce(min, [r[0] for r in ranges])
        rawindex = self.container_offset[start]
        return self.container.comments.get(rawindex, "")

    def clear_comment_ranges(self, ranges):
        c = self.container
        for start, end in ranges:
            offsets = self.calc_offsets_of_range(start, end)
            c.clear_comments(offsets)

    def iter_comments_in_segment(self):
        s = self.style[:]
        has_comments = np.where(s & style_bits.comment_bit_mask > 0)[0]
        for index in has_comments:
            rawindex = self.container_offset[index]
            yield index, self.container.comments.get(rawindex, "")

    def get_ui_name_at_index(self, index, lower_case=True):
        if lower_case:
            return "%04x" % (index + self.origin)
        else:
            return "%04X" % (index + self.origin)

    def compare_segment(self, other_segment):
        self.clear_style_bits(diff=True)
        diff = self.rawdata.data != other_segment.rawdata.data
        d = diff * np.uint8(style_bits.diff_bit_mask)
        self.style |= (diff * np.uint8(style_bits.diff_bit_mask))
        log.debug("compare_segment: # entries %d, # diffs: %d" % (len(diff), len(np.where(diff == True)[0])))

    def calc_selected_index_metadata(self, indexes):
        """Return serializable string containing style information"""
        style = self.style[indexes]
        where_comments, comments = self.get_comments_at_indexes(indexes)
        log.debug(f"after get_comments_at_indexes: {where_comments}, {comments}")
        return style, where_comments, comments

    @classmethod
    def encode_selected_index_metadata(cls, style, where_comments, comments):
        metadata = [style.tolist(), where_comments.tolist(), comments]
        j = json.dumps(metadata).encode('utf-8')
        return j

    def serialize_selected_index_metadata(self, indexes):
        a = self.calc_selected_index_metadata(indexes)
        return self.encode_selected_index_metadata(*a)

    @classmethod
    def restore_selected_index_metadata(self, encoded_meta):
        metadata = json.loads(encoded_meta.decode('utf-8'))
        style = np.asarray(metadata[0], dtype=np.uint8)
        where_comments = np.asarray(metadata[1], dtype=np.int32)
        return style, where_comments, metadata[2]


class RawSectorsSegment(Segment):
    extra_serializable_attributes = ['sector_size', 'first_sector', 'num_sectors']

    def __init__(self, container, header_size, first_sector, num_sectors, sector_size, name="Sectors", **kwargs):
        DefaultSegment.__init__(self, container, header_size, name=name, length=num_sectors*sector_size, **kwargs)
        self.sector_size = int(sector_size)
        self.first_sector = int(first_sector)
        self.num_sectors = int(num_sectors)

    def __str__(self):
        if self.num_sectors > 1:
            s = "%s (sectors %d-%d)" % (self.name, self.first_sector, self.first_sector + self.num_sectors - 1)
        else:
            s = "%s (sector %d)" % (self.name, self.first_sector)
        if self.error:
            s += " " + self.error
        return s

    @property
    def verbose_info(self):
        name = self.verbose_name or self.name
        if self.num_sectors > 1:
            s = "%s (sectors %d-%d)" % (name, self.first_sector, self.first_sector + self.num_sectors - 1)
        else:
            s = "%s (sector %d)" % (name, self.first_sector)
        s += " $%x bytes" % (len(self), )
        if self.error:
            s += "  error='%s'" % self.error
        return s

    def label(self, index, lower_case=True):
        sector, byte = divmod(index, self.sector_size)
        if lower_case:
            return "s%03d:%02x" % (sector + self.first_sector, byte)
        return "s%03d:%02X" % (sector + self.first_sector, byte)



import os

import pytest

import numpy as np
np.set_printoptions(formatter={'int': lambda x: f"{x:02x}"})

from mock import globbed_sample_atari_files

from atrip.container import Container, guess_container
from atrip.compressors import dcm
from atrip.media_type import Media, guess_media_type
from atrip.media_types.atari_disks import AtariSingleDensity

compressor = dcm.DCMCompressor()



class TestDCMBlocks:
    def setup(self):
        data = np.empty(92160, dtype=np.uint8)
        data[:] = np.repeat(np.arange(720, dtype=np.uint8), 128)
        data[::100] = 0xff
        self.byte_data = data.tobytes()
        self.container = Container(self.byte_data)
        self.media = AtariSingleDensity(self.container)
        compressor.init_packing(self.media)
        compressor.raw = compressor.pass_buffer
        compressor.index = 0

    def validate_block(self, value):
        c = compressor
        c.current[:] = c.previous[:]
        print("starting from")
        print(c.current[:128])
        code = c.get_next()
        print("using code", hex(code))
        func = c.decode_block_type_func[code]
        func(c)
        print("produced")
        print(c.current[:128])
        assert np.array_equal(c.current[:128], value)

    def test_42(self):
        c = compressor
        c.current[:124] = 0
        c.current[124] = 124
        c.current[125] = 125
        c.current[126] = 126
        c.current[127] = 127
        value = c.current[:128].copy()
        c.encode_42(-1)
        r = c.pass_buffer[0:c.pass_buffer_index]
        print(r)

        # assert len(r) == 6
        # assert r[0] == 0x42
        # assert np.array_equal(r[1:6], c.current[123:128])

        # c.previous[:] = c.current[:]
        # c.decode_42()
        # print(c.current[:128])
        # assert np.array_equal(c.current[:128], c.previous[:128])
        self.validate_block(value)

    def test_47(self):
        c = compressor
        c.current[:128] = np.arange(128, dtype=np.uint8)
        value = c.current[:128].copy()
        c.encode_47(-1)
        r = c.pass_buffer[0:c.pass_buffer_index]
        print(r)

        assert len(r) == 129
        assert r[0] == 0x47
        assert np.array_equal(r[1:129], c.current[0:128])

        # c.previous[:] = c.current[:]
        # c.current[:] = 255
        # c.decode_47()
        # print(c.current[:128])
        # assert np.array_equal(c.current[:128], c.previous[:128])
        self.validate_block(value)

    def test_41(self):
        c = compressor
        m = self.media
        c.current[:128] = m[256:384]
        c.previous[:128] = m[128:256]
        c.current[80:] = c.previous[80:]
        value = c.current[:128].copy()
        c.encode_best(False, [0x41])
        r = c.pass_buffer[0:c.pass_buffer_index]
        print("previous")
        print(c.previous[:128])
        print("current")
        print(c.current[:128])
        print("encoded")
        print(r, len(r))

        # assert len(r) == 129
        assert r[0] == 0x41
        # assert np.array_equal(r[1:129], c.current[0:128])

        # c.current[:] = c.previous[:]
        # print("starting from")
        # print(c.current[:128])
        # c.decode_41()
        # print("produced")
        # print(c.current[:128])
        # assert np.array_equal(c.current[:128], value)
        self.validate_block(value)

    def test_44(self):
        c = compressor
        m = self.media
        c.current[:128] = m[256:384]
        c.previous[:128] = m[128:256]
        c.current[:80] = c.previous[:80]
        value = c.current[:128].copy()
        c.encode_best(False, [0x44])
        r = c.pass_buffer[0:c.pass_buffer_index]
        print("previous")
        print(c.previous[:128])
        print("current")
        print(c.current[:128])
        print("encoded")
        print(r, len(r))

        # assert len(r) == 129
        assert r[0] == 0x44
        # assert np.array_equal(r[1:129], c.current[0:128])

        # c.current[:] = c.previous[:]
        # print("starting from")
        # print(c.current[:128])
        # c.decode_44()
        # print("produced")
        # print(c.current[:128])
        # assert np.array_equal(c.current[:128], value)

        self.validate_block(value)

    def test_43(self):
        c = compressor
        m = self.media
        m[256:384] = np.arange(128, dtype=np.uint8)
        c.current[:128] = m[256:384]
        c.previous[:128] = m[128:256]
        c.current[20:40] = 10
        c.current[50:60] = 99
        c.current[60:80] = 4
        value = c.current[:128].copy()
        c.encode_best(False, [0x43])
        r = c.pass_buffer[0:c.pass_buffer_index]
        print("previous")
        print(c.previous[:128])
        print("current")
        print(c.current[:128])
        print("encoded")
        print(r, len(r))

        # assert len(r) == 129
        assert r[0] == 0x43
        self.validate_block(value)

    def test_43_failure(self):
        c = compressor
        m = self.media

        # No consecutive bytes, so 43 encoding should fail
        m[256:384] = np.arange(128, dtype=np.uint8)
        c.current[:128] = m[256:384]
        c.previous[:128] = m[128:256]
        value = c.current[:128].copy()
        print("previous")
        print(c.previous[:128])
        print("current")
        print(c.current[:128])
        c.encode_best(False, [0x43])
        r = c.pass_buffer[0:c.pass_buffer_index]
        print("encoded")
        print(r, len(r))
        assert r[0] == 0x47
        self.validate_block(value)



class TestDCM:
    def setup(self):
        data = np.empty(92160, dtype=np.uint8)
        data[:] = np.repeat(np.arange(720, dtype=np.uint8), 128)
        data[::100] = 0xff
        self.byte_data = data.tobytes()
        self.container = Container(self.byte_data)
        self.media = AtariSingleDensity(self.container)

    @pytest.mark.parametrize("allowed_blocks", [
        None,
        [0x44],
        [0x41],
    ])
    def test_encode(self, allowed_blocks):
        packed = compressor.calc_packed_data(self.byte_data, self.media, allowed_blocks)
        assert packed != self.byte_data
        unpacked = compressor.calc_unpacked_data(packed)

        print(len(self.byte_data), len(packed), len(unpacked))
        orig = np.frombuffer(self.byte_data, dtype=np.uint8)
        restored = np.frombuffer(unpacked, dtype=np.uint8)
        if not np.array_equal(orig, restored):
            for i in range(720):
                start = 128 * i
                d1 = orig[start:start + 128]
                d2 = restored[start:start + 128]
                print(f"trying sector {i+1}")
                print(d1)
                print(d2)
                if not np.array_equal(d1, d2):
                    print("failure!")
                    break
        assert np.array_equal(orig, restored)


class TestFileCompression:
    @pytest.mark.parametrize(("pathname"), globbed_sample_atari_files)
    def test_glob(self, pathname):
        sample_data = np.fromfile(pathname, dtype=np.uint8)
        container = guess_container(sample_data)
        container.guess_media_type()
        m = container.media
        # can't use container._data because the container may include a header
        # which is not part of the actual media. DCM, for instance, doesn't use
        # a header while ATR does.
        packed = compressor.calc_packed_data(m.data, m)
        unpacked = compressor.calc_unpacked_data(packed)
        out = np.frombuffer(unpacked, dtype=np.uint8)
        print(len(container.media.data))
        print(len(out))
        if m.sector_size == 256 and len(m) == 183936 and len(out) == 184320:
            # DCM expands DD images full DD boot sectors, so must compare the
            # first 3 SD sectors individually
            assert np.array_equal(m.data[0:128], out[0:128])
            assert np.array_equal(m.data[128:256:], out[256:384])
            assert np.array_equal(m.data[256:384], out[512:640])
            assert np.array_equal(m.data[384:], out[768:])
        else:
            assert np.array_equal(m.data, out)

if __name__ == "__main__":
    t = TestDCMBlocks()
    t.setup()
    t.test_43()
    t.setup()
    t.test_43_failure()
    # # t.test_42()
    # # t.setup()
    # # t.test_47()
    # t.setup()
    # t.test_41()
    # t.setup()
    # t.test_44()

    # t = TestDCM()
    # t.setup()
    # t.test_encode([0x44])

    # t = TestFileCompression()
    # t.test_glob("../samples/dos_dd_test1.atr")



import numpy as np

from sawx.utils.command import Command, UndoInfo
from . import SegmentCommand
from sawx.utils.sortutil import ranges_to_indexes, indexes_to_ranges, collapse_overlapping_ranges

import logging
log = logging.getLogger(__name__)


class ClipboardCommand(SegmentCommand):
    short_name = "clipboard_command"
    ui_name = "Clipboard Abstract Command"
    serialize_order =  [
            ('segment', 'int'),
            ('blob', 'clipboard_blob')
            ]

    def __init__(self, segment, blob):
        SegmentCommand.__init__(self, segment)
        self.blob = blob

    def prepare_data(self, editor):
        pass

    def get_clipped_indexes(self, viewer):
        s = self.blob
        if s.indexes is not None:
            caret = s.dest_carets.current
            index, _ = viewer.control.table.get_index_range(*caret.rc)
            indexes = s.indexes.copy() - s.indexes[0] + index
        elif s.dest_carets.has_selection:
            ranges = collapse_overlapping_ranges(viewer.control.get_selected_ranges_including_carets(s.dest_carets))
            log.debug("ranges:", ranges)
            indexes = viewer.range_processor(ranges)
            log.debug("indexes:", indexes)
        else:
            count = len(s.data)
            ranges = []
            for c in s.dest_carets.carets:
                anchor = c.anchor_start
                if anchor[0] < 0:
                    anchor = c.rc
                index, _ = viewer.control.table.get_index_range(*anchor)
                ranges.append((index, index + count))
            ranges = collapse_overlapping_ranges(ranges)
            log.debug("ranges: {ranges}")
            indexes = viewer.range_processor(ranges)
            log.debug("indexes: {indexes}")
        max_index = len(self.segment)
        indexes = indexes[indexes < max_index]
        log.debug("indexes after limits: {str(indexes)}")
        return indexes

    def get_data(self, orig):
        data = self.blob.data
        data_len = np.alen(data)
        orig_len = np.alen(orig)
        if data_len > orig_len > 1:
            data_len = orig_len
        return data[0:data_len]

    def get_style(self, data):
        s = self.blob
        style_data = s.style
        if style_data is not None:
            style = s.style[0:np.alen(data)]
        else:
            style = None
        return style

    def do_change(self, editor, undo):
        self.prepare_data(editor)
        indexes = self.get_clipped_indexes(editor.focused_viewer)
        data = self.get_data(self.segment.data[indexes])
        log.debug("orig data: %s" % self.segment.data[indexes])
        log.debug("new data: %s" % data)
        indexes = indexes[0:np.alen(data)]
        log.debug("indexes truncated to data length: %s" % str(indexes))
        s = self.blob
        if s.relative_comment_indexes is not None:
            log.debug("relative comment indexes: %s" % (str(s.relative_comment_indexes)))
            subset = s.relative_comment_indexes[s.relative_comment_indexes < np.alen(indexes)]
            log.debug("comment index subset: %s" % str(subset))
            comment_indexes = indexes[subset]
            log.debug("new comment indexes: %s" % str(comment_indexes))
            clamped_ranges = indexes_to_ranges(indexes)
            log.debug("clamped ranges: %s" % str(clamped_ranges))
            old_comment_info = self.segment.get_comment_restore_data(clamped_ranges)
        else:
            old_comment_info = None
        undo.flags.index_range = indexes[0], indexes[-1]
        undo.flags.select_range = True
        undo.flags.byte_values_changed = True
        old_data = self.segment[indexes].copy()
        self.segment[indexes] = data
        style = self.get_style(data)
        if style is not None:
            old_style = self.segment.style[indexes].copy()
            self.segment.style[indexes] = style
        else:
            old_style = None
        if old_comment_info is not None:
            log.debug("setting comments: %s" % s.comments)
            self.segment.set_comments_at_indexes(clamped_ranges, comment_indexes, s.comments)
        return (old_data, indexes, old_style, old_comment_info)

    def undo_change(self, editor, old_data):
        old_data, old_indexes, old_style, old_comment_info = old_data
        self.segment[old_indexes] = old_data
        if old_style is not None:
            self.segment.style[old_indexes] = old_style
        if old_comment_info is not None:
            self.segment.restore_comments(old_comment_info)


class PasteCommand(ClipboardCommand):
    """Paste clipboard data at each caret or selection.

    If pasting over multiple selections will overlap, later selections will
    overwrite earlier selections.
    """
    short_name = "paste"
    ui_name = "Paste"


class PasteCommentsCommand(PasteCommand):
    """Paste comments only

    This paste command places the comments at the byte offsets of the originial
    selection. For a command that will paste comments to match the lines of a
    disassembly, see :meth:`PasteDisassemblyComments`.
    """
    short_name = "paste_comments"
    ui_name = "Paste Comments"

    def get_data(self, orig):
        return orig

    def get_style(self, data):
        None


class PasteAndRepeatCommand(PasteCommand):
    short_name = "paste_rep"
    ui_name = "Paste And Repeat"

    def get_data(self, orig):
        data = self.data
        data_len = np.alen(data)
        orig_len = np.alen(orig)
        if orig_len > data_len:
            reps = (orig_len // data_len) + 1
            data = np.tile(data, reps)
        return data[0:orig_len]


class PasteRectCommand(SegmentCommand):
    short_name = "paste_rect"
    ui_name = "Paste Rectangular"
    serialize_order =  [
            ('segment', 'int'),
            ('blob', 'clipboard_blob'),
            ]

    def __init__(self, segment, blob):
        #start_index, rows, cols, bytes_per_row, bytes):
        SegmentCommand.__init__(self, segment)
        self.blob = blob

    def __str__(self):
        s = self.blob
        return "%s @ %04x (%dx%d)" % (self.ui_name, s.dest_carets.current.index + self.segment.origin, s.num_cols, s.num_rows)

    def single_source_single_dest(self, editor, undo):
        s = self.blob
        caret = s.dest_carets.current
        i1 = caret.index
        bpr = s.dest_items_per_row
        r1, c1 = divmod(i1, bpr)
        r2 = r1 + s.num_rows
        c2 = c1 + s.num_cols
        last = r2 * bpr
        d = self.segment[:last].reshape(-1, bpr)
        r2 = min(r2, d.shape[0])
        c2 = min(c2, d.shape[1])
        undo.flags.byte_values_changed = True
        #undo.flags.index_range = i1, i2
        old_data = d[r1:r2,c1:c2].copy()
        new_data = np.fromstring(s.data, dtype=np.uint8).reshape(s.num_rows, s.num_cols)
        d[r1:r2, c1:c2] = new_data[0:r2 - r1, 0:c2 - c1]
        undo.data = (r1, c1, r2, c2, last, old_data, )
        self.undo_info = undo

    def perform(self, editor, undo):
        self.single_source_single_dest(editor, undo)

    def undo(self, editor):
        s = self.blob
        r1, c1, r2, c2, last, old_data, = self.undo_info.data
        d = self.segment[:last].reshape(-1, s.dest_items_per_row)
        d[r1:r2, c1:c2] = old_data
        return self.undo_info



import os

import numpy as np

from .. import errors
from ..segment import Segment
from ..filesystem import VTOC, Dirent, Directory, Filesystem
from .atari_dos2 import AtariDos2, AtariDos2Directory
from ..file_type import guess_file_type

class KBootDirent(Dirent):
    def __init__(self, directory):
        Dirent.__init__(self, directory, 0, 0, 0)

    def parse_raw_dirent(self):
        self.starting_sector = 4
        self.basename = self.container.basename.encode('latin1')
        if not self.basename:
            self.basename = b"KBOOT"
        elif b"." in self.basename:
            self.basename = self.basename.split(b".", 1)[0]
        if self.basename == self.basename.upper():
            self.ext = b"XEX"
        else:
            self.ext = b"xex"
        media = self.filesystem.media
        start, size = media.get_index_of_sector(4)
        i = 9
        count = media[i] + 256 * media[i+1] + 256 * 256 *media[i + 2]
        if start + count > len(media) or start + count < len(media) - 128:
            raise errors.NotEnoughSpaceOnDisk(f"KBoot header reports size {count}; media only {len(media)}")
        else:
            self.exe_size = count
            self.exe_start = start
        self.num_sectors = count // 128 + 1

    @property
    def filename(self):
        ext = (b'.' + self.ext) if self.ext else b''
        return (self.basename + ext).decode('latin1')

    @property
    def catalog_entry(self):
        return "%03d %-8s%-3s  %03d" % (4, self.basename.decode("latin1"), self.ext.decode("latin1"), self.num_sectors)

    def get_file(self):
        media = self.filesystem.media
        file_segment = guess_file_type(media, self.filename, self.exe_start, length=self.exe_size)
        self.segments = [file_segment]
        return file_segment


class KBootDirectory(AtariDos2Directory):
    def find_segment_location(self):
        media = self.media
        return 0, 0

    def calc_dirents(self):
        segments = [KBootDirent(self)]
        return segments


class KBoot(AtariDos2):
    ui_name = "Atari KBoot"
    default_executable_extension = "XEX"

    def calc_vtoc_segment(self):
        pass

    def calc_directory_segment(self):
        return KBootDirectory(self)

    @classmethod
    def create_boot_image(cls, segments, run_addr=None):
        data_segment, _ = get_xex(segments)
        payload_bytes = add_xexboot_header(data_segment.data)
        data_bytes = np.zeros(len(payload_bytes) + 16, np.uint8)
        data_bytes[16:] = payload_bytes[:]
        header_bytes = data_bytes[0:16]
        atr_header = AtrHeader(create=True)
        atr_header.check_size(len(payload_bytes))
        atr_header.encode(header_bytes)
        raw = SegmentData(data_bytes)
        atr = cls(raw, create=True)
        return atr


xexboot_header = b'\x00\x03\x00\x07\r\x07L\r\x07\x1c[\x00\x00\xa0\x00\x8c\t\x03\x8c\x04\x03\x8cD\x02\x8c\xe2\x02\x8c\xe3\x02\xc8\x84\t\x8c\x01\x03\xce\x06\x03\xa91\x8d\x00\x03\xa9R\x8d\x02\x03\xa9\x80\x8d\x08\x03\xa9\x01\x8d\x05\x03\xa9\xe3\x8d0\x02\x8d\x02\xd4\xa9\x07\x8d1\x02\x8d\x03\xd4\xa9\x00\xaa\x8d\x0b\x03\xa9\x04\x8d\n\x03 \xbc\x07\xca \xa5\x07\x85C \xa5\x07\x85D%C\xc9\xff\xf0\xf0 \xa5\x07\x85E \xa5\x07\x85F \xa5\x07\x91C\xe6C\xd0\x02\xe6D\xa5E\xc5C\xa5F\xe5D\xb0\xeb\xad\xe2\x02\r\xe3\x02\xf0\xc9\x86\x19 \xa2\x07\xa6\x19\xa0\x00\x8c\xe2\x02\x8c\xe3\x02\xf0\xb8l\xe2\x02\xad\t\x07\xd0\x0b\xad\n\x07\xd0\x03l\xe0\x02\xce\n\x07\xce\t\x07\xe0\x80\x90"\xa9@\x8d\x03\x03 Y\xe4\x10\x06\xce\x01\x07\xd0\xf1\x00\xee\n\x03\xd0\x03\xee\x0b\x03\xad\n\x03\x8d\x19\xd0\xa0\x00\xa2\x00\xbd\x00\x01\xe8`pppppF\xf8\x07p\x07ppp\x06p\x06p\x06A\xe3\x07\x00\x00\x00\x00\x00,/!$).\'\x0e\x0e\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00&2/-'


def insert_bytes(data, offset, string, color):
    s = np.frombuffer(string.upper(), dtype=np.uint8) - 32  # convert to internal
    s = s | color
    count = len(s)
    tx = offset + (20 - count) // 2
    data[tx:tx+count] = s


def add_xexboot_header(bytes, bootcode=None, title=b"DEMO", author=b"an atari user"):
    sec_size = 128
    xex_size = len(bytes)
    num_sectors = (xex_size + sec_size - 1) // sec_size
    padded_size = num_sectors * sec_size
    if xex_size < padded_size:
        bytes = np.append(bytes, np.zeros([padded_size - xex_size], dtype=np.uint8))
    paragraphs = padded_size // 16

    if bootcode is None:
        bootcode = np.copy(np.frombuffer(xexboot_header, dtype=np.uint8))
    else:
        # don't insert title or author in user supplied bootcode; would have to
        # assume that the user supplied everything desired in their own code!
        title = ""
        author = ""
    bootsize = np.alen(bootcode)
    v = bootcode[9:11].view(dtype="<u2")
    v[0] = xex_size

    bootsectors = np.zeros([384], dtype=np.uint8)
    bootsectors[0:bootsize] = bootcode

    insert_bytes(bootsectors, 268, title, 0b11000000)
    insert_bytes(bootsectors, 308, author, 0b01000000)

    image = np.append(bootsectors, bytes)
    return image



#!/usr/bin/env python

import os,sys,re,os.path,time, subprocess
from io import StringIO
from datetime import date
from optparse import OptionParser
from string import Template
from distutils.version import StrictVersion

module=None

dateformat = "%m/%d/%Y"

versionre = "([0-9]+(\.[0-9]+)+([ab][0-9]+)?)"
versionre = "(([0-9]+)\.([0-9]+)(?:\.([0-9]+))?(([ab])([0-9]+))?)"

author = "Rob McMullen <feedback@playermissile.com>"

def findLatestChangeLogVersion(options):
    fh = open("ChangeLog")
    release_date = date.today().strftime("%d %B %Y")
    version = "0.0.0"
    versions = []
    codename = ""
    for line in fh:
        match = re.match('(\d+-\d+-\d+).*',line)
        if match:
            if options.verbose: print('found date %s' % match.group(1))
            release_date = date.fromtimestamp(time.mktime(time.strptime(match.group(1),'%Y-%m-%d'))).strftime('%d %B %Y')
        match = re.match('\s+\*\s*[Rr]eleased Omnivore-([0-9]+\.[0-9]+(?:\.[0-9]+)?)',line)
        if match:
            if options.verbose: print('found version %s' % match.group(1))
            version = match.group(1)
            versions.append(version)
        release_date = None
    if release_date is None:
        release_date = date.today().strftime("%d %B %Y")
    if not versions:
        version = "0.0"
    else:
        version = versions[0]
    return version, release_date, versions

def findLatestInGit(options):
    version = StrictVersion("0.0.0")
    tags = subprocess.Popen(["git", "tag", "-l"], stdout=subprocess.PIPE).communicate()[0]
    for tag in tags.splitlines():
        match = re.match(r'%s$' % versionre, tag)
        if match:
            found = StrictVersion(match.group(1))
            if found > version:
                version = found
            if options.verbose: print("found %s, latest = %s" % (found, version))
    return str(version)

def next_version(tagged_version):
    t = tagged_version.split(".")
#    print t
    last = t[-1]
    try:
        last = str(int(last) + 1)
    except ValueError:
        for i in range(0, len(last) - 1):
            try:
                last = last[0:i+1] + str(int(last[i+1:]) + 1)
                break
            except ValueError:
                pass
    t[-1] = last
    return ".".join(t)

def getCurrentGitMD5s(tag, options):
    text = subprocess.Popen(["git", "rev-list", "%s..HEAD" % tag], stdout=subprocess.PIPE).communicate()[0]
    md5s = text.splitlines()
    return md5s

def getInitialsFromEmail(email):
    name, domain = email.split("@")
    names = name.split(".")
    initials = []
    for name in names:
        initials.append(name[0].upper())
    return "".join(initials)

def isImportantChangeLogLine(text):
    if text.startswith("Merge branch"):
        return False
    if text.startswith("updated ChangeLog & Version.py for"):
        return False
    return True

def verify_tag(tag):
    tag = str(tag)
    if tag == "HEAD":
        return tag
    verify = subprocess.Popen(["git", "tag", "-l", "%s" % (tag)], stdout=subprocess.PIPE).communicate()[0]
    if not verify:
        tag += ".0"
        verify = subprocess.Popen(["git", "tag", "-l", "%s" % (tag)], stdout=subprocess.PIPE).communicate()[0]
        if not verify:
            raise RuntimeError("tag not found! %s" % tag)
    return tag

def getGitChangeLogSuggestions(tag, options, top="HEAD"):
    tag = verify_tag(tag)
    top = verify_tag(top)
    if options.verbose: print(tag)
    suggestions = []
    text = subprocess.Popen(["git", "log", "--pretty=format:%ae--%B", "%s..%s" % (top, tag)], stdout=subprocess.PIPE).communicate()[0]
    lines = text.splitlines()
    print(lines)
    first = True
    for line in lines:
        if first:
            if "--" in line and "@" in line:
                print(line)
                email, text = line.split("--", 1)
                if isImportantChangeLogLine(text):
                    suggestions.append("* %s" % text)
                first = False
        elif not line:
            first = True
    return suggestions

def getChangeLogBlock(version, next_oldest_version, date, options):
    new_block = []
    suggestions = getGitChangeLogSuggestions(version, options, next_oldest_version)
    new_block.append("%s  %s" % (date, author))
    new_block.append("")
    if str(version) == "HEAD":
        new_block.append("...to be included in next version:")
    else:
        new_block.append("* Released Omnivore-%s" % version)
    for line in suggestions:
        new_block.append(line)
    new_block.append("")
    print("\n".join(new_block))
    return new_block

def prepend(filename, block):
    fh = open(filename)
    text = fh.read()
    fh.close()
    
    fh = open(filename, "w")
    fh.write("\n".join(block))
    fh.write("\n\n")
    fh.write(text)
    fh.close()

def replace(filename, block):
    fh = open(filename)
    current = []
    store = False
    for line in fh:
        if store:
            current.append(line)
        if not line.strip(): # skip until first blank line
            store = True
    fh.close()
    
    fh = open(filename, "w")
    fh.write("\n".join(block))
    fh.write("\n\n")
    fh.write("".join(current))
    fh.close()

def rebuild(options):
    # Command to list date of tag and tag (although newest to oldest which is
    # the opposite order than we need, hence the reversed below)
    tags = subprocess.Popen(["git", "log", "--tags", "--simplify-by-decoration", "--pretty=%ai %d"], stdout=subprocess.PIPE).communicate()[0]
    blocks = []
    next_oldest = StrictVersion("0.0.0")
    for line in reversed(tags.splitlines()):
        #day, _, _, tag = line.split()  # doesn't work; may list multiple tags
        stuff = line.split()
        day = stuff[0]
        for tag in stuff[3:]:
            if "." in tag:
                break
        tag = tag.strip("(").strip(")").strip(",")
        version = StrictVersion(tag)
        if version > next_oldest:
            if options.verbose: print("found %s" % (version))
            block = getChangeLogBlock(version, next_oldest, day, options)
            blocks[0:0] = "\n".join(block) + "\n"
            next_oldest = version
    version = "HEAD"
    day = date.today().strftime("%Y-%m-%d")
    block = getChangeLogBlock(version, next_oldest, day, options)
    print("\n".join(block) + "\n")
    
    if not options.dry_run:
        with open(options.outputfile, "w") as fh:
            for block in blocks:
                fh.write(block)

if __name__=='__main__':
    usage="usage: %prog [-m module] [-o file] [-n variablename file] [-t template] [files...]"
    parser=OptionParser(usage=usage)
    parser.add_option("-v", "--verbose", action="store_true", help="print debugging info")
    parser.add_option("--dry-run", action="store_true", help="don't actually change anything")
    parser.add_option("--version", action="store_true", help="display current version and exit")
    parser.add_option("--next-version", action="store_true", help="display what would be the next version and exit")
    parser.add_option("--rebuild", action="store_true", help="rebuild the entire ChangeLog")
    parser.add_option("-o", action="store", dest="outputfile", help="output filename", default="ChangeLog")
    (options, args) = parser.parse_args()

    tagged_version = findLatestInGit(options)
    if options.verbose:
        print("latest tagged in git: %s" % tagged_version)
    if options.version:
        print(tagged_version)
        sys.exit()
    if options.next_version:
        print(next_version(tagged_version))
        sys.exit()
    if options.rebuild:
        rebuild(options)
        sys.exit()
    
    version, latest_date, versions = findLatestChangeLogVersion(options)
    print("latest from changelog: %s" % version)
    print("all from changelog: %s" % str(versions))

    import importlib
    module = importlib.import_module("omnivore._omnivore_version")
    print(module)
    print("module version: %s" % module.version)
    
    v_changelog = StrictVersion(version)
    print(v_changelog)
    v_module = StrictVersion(module.version)
    print(v_module)
    v_tagged = StrictVersion(tagged_version)
    print(v_tagged)
    
    if v_module > v_changelog:
        print("adding to ChangeLog!")
        block = getChangeLogBlock(tagged_version, module.version, options)
        if not options.dry_run:
            prepend(options.outputfile, block)
    elif v_module == v_changelog and v_module > v_tagged:
        print("replacing ChangeLog entry!")
        block = getChangeLogBlock(tagged_version, module.version, options)
        if not options.dry_run:
            replace(options.outputfile, block)
    else:
            print("unhandled version differences...")



#!/usr/bin/env python
"""OpenGL programmable pipeline

Based on:

https://gist.githubusercontent.com/binarycrusader/5823716a1da5f0273504/raw/f8e03f4e7be917bdb4cb81a8182bdccc4c00b9d2/textured_tri.py

Texture Buffer Objects: https://www.khronos.org/opengl/wiki/Buffer_Texture
https://gist.github.com/roxlu/5090067#file-griddrawer-h-L6-L31
"""
import sys
import ctypes
import numpy as np

import OpenGL.GL as gl
from OpenGL import GLU
from OpenGL.GL import shaders
from OpenGL.arrays import vbo

import logging
log = logging.getLogger(__name__)
#log.setLevel(logging.DEBUG)


class GLProgram(object):
    def __init__(self, vertex, fragment, bindings=[]):
        vertex_shader = self.compile(vertex, gl.GL_VERTEX_SHADER)
        fragment_shader = self.compile(fragment, gl.GL_FRAGMENT_SHADER)
        self.prog = self.link([vertex_shader, fragment_shader], bindings=bindings)

    def compile(self, src, shader_type):
        shader = gl.glCreateShader(shader_type)
        gl.glShaderSource(shader, src)
        gl.glCompileShader(shader)
        result = gl.glGetShaderiv(shader, gl.GL_COMPILE_STATUS)
        if not(result):
            # TODO: this will be wrong if the user has
            # disabled traditional unpacking array support.
            raise RuntimeError(
                """Shader compile failure (%s): %s"""%(
                    result,
                    gl.glGetShaderInfoLog(shader),
                ),
                src,
                shader_type,
            )
        return shader

    def link(self, shader_objs, bindings=[]):
        prog = gl.glCreateProgram()
        for shader in shader_objs:
            gl.glAttachShader(prog, shader)
        for index, name in bindings:
            gl.glBindAttribLocation(prog, index, name)
        gl.glLinkProgram(prog)
        link_status = gl.glGetProgramiv(prog, gl.GL_LINK_STATUS)
        if link_status == gl.GL_FALSE:
            raise RuntimeError(
                """Link failure (%s): %s"""%(
                link_status,
                gl.glGetProgramInfoLog(prog),
            ))
        for shader in shader_objs:
            gl.glDeleteShader(shader)
        return prog

    def __enter__(self):
        gl.glUseProgram(self.prog)

    def __exit__(self, type, value, traceback):
        #gl.glBindVertexArray(0)
        gl.glUseProgram(0)



vertexShader = """
#version 120

void main()
{
    gl_TexCoord[0] = gl_MultiTexCoord0;
    gl_Position = ftransform();
}
"""

fragmentShader = """
#version 120

uniform sampler1D palette;
uniform sampler2D tex;

void main()
{
    vec4 source;
    vec4 pcolor;

    source = texture2D(tex, gl_TexCoord[0].st);
    //gl_FragColor = vec4(gl_TexCoord[0].st, 0, 255);
    pcolor = texture1D(palette, source.r);
    gl_FragColor = (pcolor * 0.5) + (source * 0.5);
    gl_FragColor = pcolor;
    //gl_FragColor = vec4(1,0,0,1) + (pcolor * 0.001);
    //gl_FragColor = vec4(gl_TexCoord[0].st,0,1) + (pcolor * 0.001);
}
"""


class GLSLTextureCanvas(object):
    def __init__(self, initial_palette):
        self.ui_init_context()
        self.init_attributes()
        self.ui_bind_events()
        self.set_palette_data(initial_palette)

    def ui_init_context(self):
        """Initialize the GL context from the UI toolkit"""
        raise NotImplementedError

    def ui_bind_events(self):
        """Bind any event handlers that the UI toolkit needs for repainting or
        updating the display."""
        raise NotImplementedError

    def ui_set_current(self):
        """Set the current GLContext using the UI toolkit-specific code, using
        the context created by ui_init_context"""
        raise NotImplementedError

    def ui_swap_buffers(self):
        """Swaps the OpenGL back and front buffers for double buffering"""
        raise NotImplementedError

    def ui_get_window_size(self):
        """Return the width and height of the drawable area (sometimes referred
        to as the client area) of the window"""
        raise NotImplementedError

    def init_attributes(self):
        self.vertex_data = np.array([
            # X,  Y,   U, V
            (-1, -1,   0, 0),
            ( 1, -1,   1, 0),
            (-1,  1,   0, 1),
            ( 1,  1,   1, 1)
        ], dtype=np.float32)
        self.shader_prog = None
        self.vao_id = None
        self.vbo_id = None
        self.display_texture = None
        self.tex_uniform = None

        self.finished_init = False

    def init_shader(self):
        self.shader_prog = GLProgram(vertexShader, fragmentShader, [(0, "position"), (1, "in_tex_coords")])

        # load texture and assign texture unit for shaders
        self.tex_uniform = gl.glGetUniformLocation(self.shader_prog.prog, 'tex')
        self.palette_uniform = gl.glGetUniformLocation(self.shader_prog.prog, 'palette')

        # # load texture and assign texture unit for shaders
        # self.tex_uniform = gl.glGetUniformLocation(self.shader_prog.prog, 'tex')
        # self.palette_uniform = gl.glGetUniformLocation(self.shader_prog.prog, 'palette')

    def init_context(self):
        """init the texture - this has to happen after an OpenGL context
        has been created
        """

        # make the OpenGL context associated with this canvas the current one
        self.ui_set_current()

        self.init_shader()

        # Core OpenGL requires that at least one OpenGL vertex array be bound
        #self.vao_id = gl.glGenVertexArrays(1)
        #gl.glBindVertexArray(self.vao_id)

        # Need self.vbo_id for triangle vertices and texture UV coordinates
        self.vbo_id = gl.glGenBuffers(1)
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, self.vbo_id)
        gl.glBufferData(gl.GL_ARRAY_BUFFER, self.vertex_data.nbytes, self.vertex_data,
            gl.GL_STATIC_DRAW)

        # load texture and assign texture unit for shaders
        self.display_texture = self.load_texture()

        self.palette_texture = self.load_palette()
        # # load palette and assign texture unit for shaders
        # self.palette_id = gl.glGenBuffers(1)
        # gl.glBindBuffer(gl.GL_TEXTURE_BUFFER, self.palette_id)
        # gl.glBufferData(gl.GL_TEXTURE_BUFFER, NTSC.nbytes, NTSC,
        #     gl.GL_STATIC_DRAW)
        # gl.glTexBuffer(gl.GL_TEXTURE_BUFFER, gl.GL_RGBA8, self.palette_id)
        # self.palette_texture = gl.glGenTextures(1)

        #gl.glEnableVertexAttribArray(0)
        #gl.glEnableVertexAttribArray(1)
        #gl.glVertexAttribPointer(0, 2, gl.GL_FLOAT, gl.GL_FALSE, 16, None)
        # the last parameter is a pointer
        #gl.glVertexAttribPointer(1, 2, gl.GL_FLOAT, gl.GL_TRUE, 16, ctypes.c_void_p(8))

        # Finished
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, 0)
        #gl.glBindVertexArray(0)

        self.finished_init = True

    def get_color_indexed_texture_data(self, raw=None):
        if raw is None:
            w = 16
            h = 16
            raw = np.empty((256,), dtype=np.uint8)
            raw[:] = np.arange(256, dtype=np.uint8)
            raw = raw.reshape((h, w))
        return raw

    def get_rgba_texture_data(self, raw=None):
        raw = self.get_color_indexed_texture_data(raw)
        h, w = raw.shape
        data = np.empty((h * w, 4), dtype=np.uint8)
        src = raw.reshape((h * w))
        data[:,0] = src[:]
        data[:,1] = src[:]
        data[:,2] = src[:]
        data[:,3] = 255
        return data.reshape((h, w, 4))

    def load_texture(self, filename=None):
        if filename is not None or False:
            from PIL import Image
            filename = "flicky.png"
            image = Image.open(filename)
            ix = image.size[0]
            iy = image.size[1]
            log.debug("image: %s, %dx%d" % (filename, ix, iy))
            w, h = image.size
            rgba = 4
            data = image.tobytes("raw", "RGBX", 0, -1)
        else:
            data = self.get_rgba_texture_data(-1)
            h, w, rgba = data.shape

        # generate a texture id, make it current
        texture = gl.glGenTextures(1)
        gl.glBindTexture(gl.GL_TEXTURE_2D, texture)
        log.debug("load_texture: glGenTextures returns %d, shape=%s" % (texture, data.shape))
        if texture is None:
            sys.exit()

        # texture mode and parameters controlling wrapping and scaling
        gl.glTexEnvf(gl.GL_TEXTURE_ENV, gl.GL_TEXTURE_ENV_MODE, gl.GL_REPLACE)
        gl.glTexParameterf(
            gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_S, gl.GL_REPEAT)
        gl.glTexParameterf(
            gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_T, gl.GL_REPEAT)
        gl.glTexParameterf(
            gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_NEAREST)
        gl.glTexParameterf(
            gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_NEAREST)
        gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, rgba, w, h, 0,
            gl.GL_RGBA, gl.GL_UNSIGNED_BYTE, data)
        # gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, 0, 0, w, h,
        #                 gl.GL_RGBA, gl.GL_UNSIGNED_BYTE, data)

        self.update_texture(texture, data)
        return texture

    def update_texture(self, texture, data):
        # map the image data to the texture. note that if the input
        # type is GL_FLOAT, the values must be in the range [0..1]
        log.debug("update_texture: texture=%d, shape=%s" % (texture, data.shape))
        # if True:
        #     return
        gl.glBindTexture(gl.GL_TEXTURE_2D, texture)
        h, w, rgba = data.shape
        gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, rgba, w, h, 0,
            gl.GL_RGBA, gl.GL_UNSIGNED_BYTE, data)
        # gl.glTexSubImage2D(gl.GL_TEXTURE_2D, 0, 0, 0, w, h,
        #     gl.GL_RGBA, gl.GL_UNSIGNED_BYTE, data)
        #gl.glBindTexture(gl.GL_TEXTURE_2D, 0)

    def set_palette_data(self, data):
        self.palette_data = data
        if self.finished_init:
            self.update_palette()

    def load_palette(self):
        # generate a texture id, make it current
        texture = gl.glGenTextures(1)
        gl.glBindTexture(gl.GL_TEXTURE_1D, texture)
        log.debug("load_palette: texture=%d, shape=%s" % (texture, self.palette_data.shape))
        if texture is None:
            sys.exit()
        entries, rgba = self.palette_data.shape

        # texture mode and parameters controlling wrapping and scaling
        gl.glTexEnvf(gl.GL_TEXTURE_ENV, gl.GL_TEXTURE_ENV_MODE, gl.GL_REPLACE)
        gl.glTexParameterf(
            gl.GL_TEXTURE_1D, gl.GL_TEXTURE_WRAP_S, gl.GL_CLAMP)
        gl.glTexParameterf(
            gl.GL_TEXTURE_1D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_NEAREST)
        gl.glTexParameterf(
            gl.GL_TEXTURE_1D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_NEAREST)
        gl.glTexImage1D(gl.GL_TEXTURE_1D, 0, rgba, entries, 0,
            gl.GL_RGBA, gl.GL_UNSIGNED_BYTE, self.palette_data)
        return texture

    def update_palette(self):
        """Update the palette colors"""
        raise NotImplementedError

    def on_size(self, event):
        """called when window is repainted """
        self.set_aspect()

        if self.finished_init:
            self.on_draw()

    def set_aspect(self):
        if not self.finished_init:
            return
        w, h = self.ui_get_window_size()
        a = (240.0 / 336.0) * w / h
        xspan = 1
        yspan = 1
        if a > 1:
            xspan *= a
        else:
            yspan = xspan/a

        gl.glViewport(0, 0, w, h);
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadIdentity()
        gl.glOrtho(-1*xspan, xspan, -1*yspan, yspan, -1, 1)
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glLoadIdentity()

    def on_draw(self):
        """draw function """

        # make the OpenGL context associated with this canvas the current one
        self.ui_set_current()

        self.render()

        # swap the front and back buffers so that the texture is visible
        self.ui_swap_buffers()

    def render(self):
        gl.glClearColor(0, 0, 0, 1)
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
        gl.glEnable(gl.GL_TEXTURE_1D)
        gl.glEnable(gl.GL_TEXTURE_2D)
        gl.glDisable(gl.GL_LIGHTING)
        gl.glDisable(gl.GL_CULL_FACE)
        gl.glColor(1.0, 1.0, 1.0, 1.0)

        with self.shader_prog:
            # Activate array
            #gl.glBindVertexArray(self.vao_id)

            # FIXME: Why does the following work? tex_uniform is 1,
            # palette_uniform is 0, but I have to set the uniform for
            # tex_uniform to 0 and the uniform for palette_uniform to 1.
            # Obviously I'm not understanding something.
            #
            # See: http://stackoverflow.com/questions/26622204
            # https://www.opengl.org/discussion_boards/showthread.php/174926-when-to-use-glActiveTexture

            # Activate texture
            #print self.tex_uniform, self.palette_uniform, self.display_texture, getattr(self, 'palette_id') if hasattr(self, 'palette_id') else "None"
            gl.glActiveTexture(gl.GL_TEXTURE0 + self.tex_uniform)
            gl.glBindTexture(gl.GL_TEXTURE_2D, self.display_texture)
            gl.glUniform1i(self.tex_uniform, 1)

            # Activate palette texture
            gl.glActiveTexture(gl.GL_TEXTURE0 + self.palette_uniform)
            if hasattr(self, 'palette_id'):
                gl.glBindTexture(gl.GL_TEXTURE_BUFFER, self.palette_texture)
                gl.glTexBuffer(gl.GL_TEXTURE_BUFFER, gl.GL_RGBA8, self.palette_id)
            else:
                gl.glBindTexture(gl.GL_TEXTURE_1D, self.palette_texture)

            gl.glUniform1i(self.palette_uniform, 0)

            # # Activate array
            # gl.glBindVertexArray(self.vao_id)
            #gl.glEnableClientState(gl.GL_VERTEX_ARRAY)  # FIXME: deprecated
            #gl.glVertexPointer(2, gl.GL_FLOAT, 0, None)  # FIXME: deprecated

            # draw triangle
            #gl.glDrawArrays(gl.GL_TRIANGLE_STRIP, 0, 4)
            gl.glBindTexture(gl.GL_TEXTURE_2D, self.display_texture)
            gl.glBegin(gl.GL_TRIANGLE_STRIP)
            for pt in self.vertex_data:
                gl.glTexCoord2f(pt[2], pt[3])
                gl.glVertex2f(pt[0], pt[1])
            gl.glEnd()


class LegacyTextureCanvas(GLSLTextureCanvas):
    def init_shader(self):
        pass

    def get_rgba_texture_data(self, raw):
        raise NotImplementedError("subclass must return shape (w,h,4) array")

    def render(self):
        gl.glClearColor(0, 0, 0, 1)
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)

        self.set_aspect()

        gl.glDisable(gl.GL_LIGHTING)
        # Don't cull polygons that are wound the wrong way.
        gl.glDisable(gl.GL_CULL_FACE)

        gl.glEnable(gl.GL_TEXTURE_2D)
        gl.glColor(1.0, 1.0, 1.0, 1.0)

        gl.glBindTexture(gl.GL_TEXTURE_2D, self.display_texture)
        gl.glBegin(gl.GL_TRIANGLE_STRIP)
        for pt in self.vertex_data:
            gl.glTexCoord2f(pt[2], pt[3])
            gl.glVertex2f(pt[0], pt[1])
        gl.glEnd()



import os

import pytest

import numpy as np

from atrip import errors
from atrip.stringifier import find_stringifier_by_name

stringifiers = [
    ('repr', "\\x00\\x01"),
    ('hexify', "00 01 02 03"),
    ('c_bytes', "0x00,0x01,0x02,0x03"),
    ('basic_data', "DATA 0,1,2,3"),
]

class TestStringifier:
    def setup(self):
        data = np.arange(128, dtype=np.uint8)
        self.byte_data = data.tobytes()

    @pytest.mark.parametrize("name,text_check", stringifiers)
    def test_stringifier(self, name, text_check):
        s = find_stringifier_by_name(name)

        text = s.calc_text(self.byte_data)
        print(text)
        assert text != self.byte_data
        c = len(text_check)
        assert text[0:c] == text_check

        with pytest.raises(errors.UnsupportedAlgorithm):
            parsed = s.calc_byte_data(text)
            print(len(self.byte_data), len(parsed))
            assert parsed == self.byte_data

if __name__ == "__main__":
    t = TestStringifier()
    t.setup()
    t.test_stringifier('hexify', "00 01")
    t.test_stringifier('basic_data', "DATA 0,1,2")
    t.test_stringifier('repr', '\\x00\\x01')
    t.test_stringifier('c_bytes', "0x00,0x01,0x02,0x03")



import os
import sys
import subprocess
from setuptools import setup, find_packages, Extension
import glob
from distutils.command.clean import clean
try:
    import numpy as np
except ImportError:
    sys.exit('Please install numpy first, e.g. "pip install numpy"')

if sys.version_info < (3, 6):
    sys.exit('Omnivore requires Python 3.6 or higher')

try:
    from Cython.Build import cythonize
except ImportError:
    def cythonize(args):
        missing = []
        for ext in args:
            for src in ext.sources:
                if src.endswith(".pyx"):
                    gen = src[:-4] + ".c"
                    if not os.path.exists(gen):
                        missing.append(gen)
        if missing:
            print("Cython is required because this version was not shipped with the cythonized C files:")
            print("\n".join(missing))
            sys.exit(1)
        return args


if sys.platform.startswith("win"):
    extra_compile_args = ["-DMSVC", "-D_CRT_SECURE_NO_WARNINGS", "/Zi"]
    extra_link_args=['/DEBUG']
    libatari800_config_include = "libatari800/include/win"
else:
    extra_compile_args = ["-g"]
    #extra_compile_args = ["-Og"]
    #extra_compile_args = ["-O3"]
    extra_link_args = []
    libatari800_config_include = "libatari800/include/linux"

# extra_compile_args.append("-DDEBUG_THREAD_STATUS")
# extra_compile_args.append("-DDEBUG_REGISTER_CALLBACK")
# extra_compile_args.append("-DDEBUG_BREAKPOINT")
# extra_compile_args.append("-DDEBUG_POSTFIX_STACK")


class clean_py(clean):
    def run(self):
        clean.run(self)
        files = [
            "omnivore/*/*.so",
            "omnivore/*/*.pyd",
            "omnivore/*/*/*.so",
            "omnivore/*/*/*.pyd",
            "omnivore/arch/antic_speedups.c",
            "omnivore/disassembler/cputables.py",
            "lib6502/lib6502.c",
            "libatari800/libatari800.c",
            "libmac65/libmac65.c",
            "libudis/declarations.pyx",
            "libudis/declarations.c",
            "libudis/libudis.c",
            "libudis/parse_udis_cpu.c",
            "libudis/stringify_udis_cpu.c",
            "libudis/stringify_udis_cpu.h",
        ]
        for pathspec in files:
            for path in glob.glob(pathspec):
                print("cleaning" + str(path))
                try:
                    os.unlink(path)
                except OSError:
                    pass

if "clean" in sys.argv:
    # prevent extensions from being 
    ext_modules = []
else:
    cputables_path = "atrip/disassemblers/cputables.py"
    parse_gen_path = "libudis/parse_udis_cpu.c"
    if not os.path.exists(cputables_path):
        subprocess.run([sys.executable, 'libudis/cpugen.py'])
    out = {}
    exec(compile(open(cputables_path).read(), cputables_path, 'exec'), out)
    processors = out['processors']
    if '6502' not in processors:
        for path in [cputables_path, parse_gen_path]:
            try:
                os.unlink(path)
            except IOError:
                pass
        sys.exit("Disassemblers not generated; try 'git submodule init; git submodule update' and rerun.")

    if not os.path.exists(parse_gen_path):
        subprocess.run([sys.executable, 'libudis/parse_gen.py'])

    extensions = [
        Extension("omnivore.arch.antic_speedups",
                  sources=["omnivore/arch/antic_speedups.pyx"],
                  extra_compile_args = extra_compile_args,
                  extra_link_args = extra_link_args,
                  include_dirs = [np.get_include()],
                  ),
        Extension("omnivore.arch.pixel_speedups",
                  sources=["omnivore/arch/pixel_speedups.pyx"],
                  extra_compile_args = extra_compile_args,
                  extra_link_args = extra_link_args,
                  include_dirs = [np.get_include()],
                  ),
        Extension("omnivore.emulators.atari8bit.libatari800",
            sources = [
                "libatari800/atari800_bridge.c",
                "libatari800/atari800_antic.c",
                "libatari800/atari800_cpu.c",
                "libatari800/libatari800.pyx",
                "libatari800/tinycthread.c",
                "libatari800/atari800/src/libatari800/main.c",
                "libatari800/atari800/src/libatari800/input.c",
                "libatari800/atari800/src/libatari800/video.c",
                "libatari800/atari800/src/libatari800/init.c",
                "libatari800/atari800/src/libatari800/sound.c",
                "libatari800/atari800/src/libatari800/statesav.c",
                "libatari800/atari800/src/afile.c",
                # "libatari800/atari800/src/antic.c",
                "libatari800/atari800/src/atari.c",
                "libatari800/atari800/src/binload.c",
                "libatari800/atari800/src/cartridge.c",
                "libatari800/atari800/src/cassette.c",
                "libatari800/atari800/src/compfile.c",
                "libatari800/atari800/src/cfg.c",
                # "libatari800/atari800/src/cpu.c",
                "libatari800/atari800/src/crc32.c",
                "libatari800/atari800/src/devices.c",
                "libatari800/atari800/src/esc.c",
                "libatari800/atari800/src/gtia.c",
                "libatari800/atari800/src/img_tape.c",
                "libatari800/atari800/src/log.c",
                "libatari800/atari800/src/memory.c",
                "libatari800/atari800/src/monitor.c",
                "libatari800/atari800/src/pbi.c",
                "libatari800/atari800/src/pia.c",
                "libatari800/atari800/src/pokey.c",
                "libatari800/atari800/src/rtime.c",
                "libatari800/atari800/src/sio.c",
                "libatari800/atari800/src/sysrom.c",
                "libatari800/atari800/src/util.c",
                "libatari800/atari800/src/input.c",
                "libatari800/atari800/src/statesav.c",
                "libatari800/atari800/src/artifact.c",
                "libatari800/atari800/src/colours.c",
                "libatari800/atari800/src/colours_ntsc.c",
                "libatari800/atari800/src/colours_pal.c",
                "libatari800/atari800/src/colours_external.c",
                "libatari800/atari800/src/screen.c",
                "libatari800/atari800/src/cycle_map.c",
                "libatari800/atari800/src/pbi_mio.c",
                "libatari800/atari800/src/pbi_bb.c",
                "libatari800/atari800/src/pbi_scsi.c",
                "libatari800/atari800/src/pokeysnd.c",
                "libatari800/atari800/src/mzpokeysnd.c",
                "libatari800/atari800/src/remez.c",
                "libatari800/atari800/src/sndsave.c",
                "libatari800/atari800/src/sound.c",
                "libatari800/atari800/src/pbi_xld.c",
                "libatari800/atari800/src/voicebox.c",
                "libatari800/atari800/src/votrax.c",
                "libatari800/atari800/src/votraxsnd.c",
                "libatari800/atari800/src/roms/altirra_5200_os.c",
                "libatari800/atari800/src/roms/altirraos_800.c",
                "libatari800/atari800/src/roms/altirraos_xl.c",
                "libatari800/atari800/src/roms/altirra_basic.c",
                "libdebugger/libdebugger.c",
                "libudis/history.c",
            ],
            depends = [
                "libatari800/atari800_bridge.h",
                "libatari800/tinycthread.h",
                "libatari800/include/linux/config.h",
                "libatari800/include/win/config.h",
                "libatari800/include/win/dirent.h",
                "libatari800/atari800/src/libatari800/libatari800.h",
                "libatari800/atari800/src/libatari800/main.h",
                "libatari800/atari800/src/libatari800/input.h",
                "libatari800/atari800/src/libatari800/video.h",
                "libatari800/atari800/src/libatari800/init.h",
                "libatari800/atari800/src/libatari800/sound.h",
                "libatari800/atari800/src/libatari800/statesav.h",
                "libatari800/atari800/src/afile.h",
                "libatari800/atari800/src/antic.h",
                "libatari800/atari800/src/atari.h",
                "libatari800/atari800/src/binload.h",
                "libatari800/atari800/src/cartridge.h",
                "libatari800/atari800/src/cassette.h",
                "libatari800/atari800/src/compfile.h",
                "libatari800/atari800/src/cfg.h",
                "libatari800/atari800/src/cpu.h",
                "libatari800/atari800/src/crc32.h",
                "libatari800/atari800/src/devices.h",
                "libatari800/atari800/src/esc.h",
                "libatari800/atari800/src/gtia.h",
                "libatari800/atari800/src/img_tape.h",
                "libatari800/atari800/src/log.h",
                "libatari800/atari800/src/memory.h",
                "libatari800/atari800/src/monitor.h",
                "libatari800/atari800/src/pbi.h",
                "libatari800/atari800/src/pia.h",
                "libatari800/atari800/src/pokey.h",
                "libatari800/atari800/src/rtime.h",
                "libatari800/atari800/src/sio.h",
                "libatari800/atari800/src/sysrom.h",
                "libatari800/atari800/src/util.h",
                "libatari800/atari800/src/input.h",
                "libatari800/atari800/src/statesav.h",
                "libatari800/atari800/src/ui.h",
                "libatari800/atari800/src/artifact.h",
                "libatari800/atari800/src/colours.h",
                "libatari800/atari800/src/colours_ntsc.h",
                "libatari800/atari800/src/colours_pal.h",
                "libatari800/atari800/src/colours_external.h",
                "libatari800/atari800/src/screen.h",
                "libatari800/atari800/src/cycle_map.h",
                "libatari800/atari800/src/pbi_mio.h",
                "libatari800/atari800/src/pbi_bb.h",
                "libatari800/atari800/src/pbi_scsi.h",
                "libatari800/atari800/src/pokeysnd.h",
                "libatari800/atari800/src/mzpokeysnd.h",
                "libatari800/atari800/src/remez.h",
                "libatari800/atari800/src/sndsave.h",
                "libatari800/atari800/src/sound.h",
                "libatari800/atari800/src/pbi_xld.h",
                "libatari800/atari800/src/voicebox.h",
                "libatari800/atari800/src/votrax.h",
                "libatari800/atari800/src/votraxsnd.h",
                "libatari800/atari800/src/roms/altirra_5200_os.h",
                "libatari800/atari800/src/roms/altirraos_800.h",
                "libatari800/atari800/src/roms/altirraos_xl.h",
                "libatari800/atari800/src/roms/altirra_basic.h",
                "libdebugger/libdebugger.h",
                "libudis/libudis.h",
                "libudis/libudis_flags.h",
            ],
            extra_compile_args = extra_compile_args,
            extra_link_args = extra_link_args,
            include_dirs = [libatari800_config_include, "libatari800/atari800/src", "libatari800/atari800/src/libatari800", "libdebugger", "libudis", np.get_include()],
            undef_macros = [ "NDEBUG" ],
        ),
        Extension("omnivore.emulators.generic6502.lib6502",
            sources = [
                "lib6502/lib6502.pyx",
                "lib6502/6502-emu_wrapper.c",
                "lib6502/libcrabapple.c",
                "lib6502/6502-emu/6502.c",
                "libdebugger/libdebugger.c",
                "libudis/history.c",
            ],
            depends = [
                "lib6502/6502-emu_wrapper.h",
                "lib6502/libcrabapple.h",
                "lib6502/6502-emu/6502.h",
                "libdebugger/libdebugger.h",
                "libudis/libudis.h",
                "libudis/libudis_flags.h",
            ],
            extra_compile_args = extra_compile_args,
            extra_link_args = extra_link_args,
            include_dirs = ["lib6502", "lib6502/6502-emu", "libdebugger", "libudis", np.get_include()],
            undef_macros = [ "NDEBUG" ],
        ),
        Extension("atrip.disassemblers.libudis",
            sources = [
                "libudis/libudis.pyx",
                "libudis/parse_udis_cpu.c",
                "libudis/parse_custom.c",
                "libudis/stringify_udis_cpu.c",
                "libudis/stringify_custom.c",
                "libudis/history.c",
            ],
            depends = [
                "libudis/libudis.h",
                "libudis/libudis_flags.h",
            ],
            extra_compile_args = extra_compile_args,
            include_dirs = ["libdebugger", "libudis", np.get_include()],
            ),
        Extension("atrip.assemblers.libmac65",
            sources = [
                "libmac65/libmac65.pyx",
                "libmac65/cython_interface.c",
                "libmac65/atasm/src/asm.c",
                "libmac65/atasm/src/symbol.c",
                "libmac65/atasm/src/parser.c",
                "libmac65/atasm/src/setparse.c",
                "libmac65/atasm/src/inc_path.c",
                "libmac65/atasm/src/crc32.c",
                "libmac65/atasm/src/atasm_err.c",
                ],
            depends = [
                "libmac65/atasm/src/atasm_err.h",
                "libmac65/atasm/src/compat.h",
                "libmac65/atasm/src/directive.h",
                "libmac65/atasm/src/inc_path.h",
                "libmac65/atasm/src/ops.h",
                "libmac65/atasm/src/symbol.h",
                ],
          extra_compile_args = extra_compile_args,
          include_dirs = ["libmac65/atasm/src"],
          )
    ]
    ext_modules = cythonize(extensions)

exec(compile(open('omnivore/_version.py').read(), 'omnivore/_version.py', 'exec'))
exec(compile(open('omnivore/_metadata.py').read(), 'omnivore/_metadata.py', 'exec'))

setup(
    name = "omnivore",
    version = __version__,
    author = __author__,
    author_email = __author_email__,
    url = __url__,
    classifiers = [c.strip() for c in """\
        Development Status :: 5 - Production/Stable
        Intended Audience :: Developers
        License :: OSI Approved :: GNU General Public License (GPL)
        Operating System :: MacOS
        Operating System :: Microsoft :: Windows
        Operating System :: OS Independent
        Operating System :: POSIX
        Operating System :: Unix
        Programming Language :: Python
        Topic :: Utilities
        Topic :: Software Development :: Assemblers
    """.splitlines() if len(c.strip()) > 0],
    python_requires = ">=3.6",
    description = "Unified front-end providing common interface to control several 8 bit emulators",
    long_description = open('README.rst').read(),
    ext_modules = ext_modules,
    packages = find_packages(exclude=["libudis"]),
    include_package_data=True,
    scripts = [
        "scripts/atrip",
        "scripts/omnivore",
        "scripts/omnimu",
    ],
    entry_points={
        "atrip.archivers": [
            'zip = atrip.archivers.zip',
            'tar = atrip.archivers.tar',
        ],

        "atrip.assemblers": [
            'mac65 = atrip.assemblers.mac65',
        ],

        "atrip.compressors": [
            'zlib = atrip.compressors.zlib',
            'gzip = atrip.compressors.gzip',
            'bzip = atrip.compressors.bzip',
            'lzma = atrip.compressors.lzma',
            'Z = atrip.compressors.unix_compress',
            'lz4 = atrip.compressors.lz4',
            'dcm = atrip.compressors.dcm',
        ],

        "atrip.stringifiers": [
            'hexify = atrip.stringifiers.hexify',
            'c_bytes = atrip.stringifiers.c_bytes',
            'basic_data = atrip.stringifiers.basic_data',
        ],

        "atrip.media_types": [
            'atari_disks = atrip.media_types.atari_disks',
            'atari_carts = atrip.media_types.atari_carts',
            'atari_tapes = atrip.media_types.atari_tapes',
            'apple_disks = atrip.media_types.apple_disks',
        ],

        "atrip.filesystems": [
            'atari_dos = atrip.filesystems.atari_dos2',
            'atari_cas = atrip.filesystems.atari_cas',
            'atari_jumpman = atrip.filesystems.atari_jumpman',
            'kboot = atrip.filesystems.kboot',
            'apple_dos33 = atrip.filesystems.apple_dos33',
        ],

        "atrip.file_types": [
            'atari_xex = atrip.file_types.atari_xex',
            'apple_bsave = atrip.file_types.apple_dos33_bsave',
        ],

        "atrip.signatures": [
            'atari2600_cart = atrip.signatures.atari2600_cart',
            'atari2600_starpath = atrip.signatures.atari2600_starpath',
            'atari5200_cart = atrip.signatures.atari5200_cart',
            'atari800_cart = atrip.signatures.atari800_cart',
            'vectrex = atrip.signatures.vectrex',
        ],

        "omnivore.emulators": [
            'atari8bit = omnivore.emulators.atari8bit.atari800',
            'apple2 = omnivore.emulators.apple2.crabapple',
            'generic6502 = omnivore.emulators.generic6502.generic6502',
        ],

        "omnivore.viewers": [
            'bitmap = omnivore.viewers.bitmap',
            'pixelmap = omnivore.viewers.pixelmap',
            'char = omnivore.viewers.char',
            'hex = omnivore.viewers.hex',
            'info = omnivore.viewers.info',
            'map = omnivore.viewers.map',
            'jumpman = omnivore.viewers.jumpman',
            'emulator = omnivore.viewers.emulator',
            'apple2 = omnivore.viewers.apple2',
            'memory = omnivore.viewers.memory',
            'skeleton = omnivore.viewers.skeleton',
            'disasm = omnivore.viewers.disasm',
            'history = omnivore.viewers.history',
            'labels = omnivore.viewers.labels',
        ],

        "sawx.documents": [
            '00byte = omnivore.document',
            '99text = sawx.documents.text',
        ],

        "sawx.editors": [
            '00byte = omnivore.editors.byte_editor',
            '00emu = omnivore.editors.emulation_editor',
            'html = sawx.editors.html_viewer',
            'text = sawx.editors.text_editor',
        ],

        "sawx.loaders": [
            '00atrip = atrip.omnivore_loader',
            '99fleep = sawx.loaders.fleep',
            '98text = sawx.loaders.text',
        ],

        "sawx.remember": [
            'fonts = omnivore.arch.fonts',
            'app = sawx.application',
        ],
    },
    platforms = ["Windows", "Linux", "Mac OS-X", "Unix"],
    zip_safe = False,
    install_requires = [
    'pathlib2',  # pip install processes from the top down!
    'numpy',

    'python-slugify',
    'ply',
    'lz4',
    'construct<2.9',  # Construct 2.9 changed the String class
    'pytz',
    'pyparsing',
    'configobj',
    'bson<1.0.0',
    'jsonpickle',
    # 'pyopengl-accelerate',  # not required, and a pain on some platform/os combos
    'pyopengl',
    'fleep',  # not actually required
    'appdirs',
    'wxpython',
    'pillow',
    'six',

    'numpy',
    'pathlib2',  # python setup.py develop processes from the bottom up!
    ],
    cmdclass = {"clean": clean_py},
)



import os
import sys

import wx

from ..ui.segment_grid import SegmentGridControl, SegmentTable
from sawx.ui import compactgrid as cg
# from ..arch.disasm import get_style_name

from ..viewer import SegmentViewer

import logging
log = logging.getLogger(__name__)


hex_digits_on_keypad=[
    wx.WXK_NUMPAD0, wx.WXK_NUMPAD1, wx.WXK_NUMPAD2, wx.WXK_NUMPAD3,
    wx.WXK_NUMPAD4, wx.WXK_NUMPAD5, wx.WXK_NUMPAD6, wx.WXK_NUMPAD7,
    wx.WXK_NUMPAD8, wx.WXK_NUMPAD9
    ]

def is_valid_hex_digit(key):
    return key in hex_digits_on_keypad or (key>=ord('0') and key<=ord('9')) or (key>=ord('A') and key<=ord('F')) or (key>=ord('a') and key<=ord('f'))

def get_valid_hex_digit(key):
    if key in hex_digits_on_keypad:
        return chr(ord('0') + key - wx.WXK_NUMPAD0)
    elif (key>=ord('0') and key<=ord('9')) or (key>=ord('A') and key<=ord('F')) or (key>=ord('a') and key<=ord('f')):
        return chr(key)
    else:
        return None


class HexTextCtrl(cg.GridCellTextCtrl):
    def __init__(self, parent, id, *args, **kwargs):
        # Don't use the validator here, because apparently we can't
        # reset the validator based on the columns.  We have to do the
        # validation ourselves using EVT_KEY_DOWN.
        super().__init__(parent, id, 2, *args, **kwargs)

    def is_valid_keycode(self, keycode):
        return is_valid_hex_digit(keycode)

    def get_processed_value(self):
        text = self.GetValue()
        print(("text value: %s" % text))
        return int(text, 16)


class HexEditControl(SegmentGridControl):
    """
    View for editing in hexidecimal notation.
    """
    short_name = "hex"

    def set_viewer_defaults(self):
        self.items_per_row = self.view_params.hex_grid_width

    def verify_keycode_can_start_edit(self, c):
        return get_valid_hex_digit(c)

    def create_hidden_text_ctrl(self):
        c = HexTextCtrl(self, -1, pos=(600,100), size=(400,24))
        return c


class HexEditViewer(SegmentViewer):
    name = "hex"

    ui_name = "Hex"

    control_cls = HexEditControl

    has_hex = True

    has_width = True

    def recalc_data_model(self):
        """Rebuild the data model after a document formatting (or other
        structural change) or loading a new document.
        """
        self.control.recalc_view()

    def update_carets(self, flags):
        pass

    def select_hgr(self):
        from sawx.ui.compactgrid_mouse import Caret
        ch = self.control.caret_handler
        ch.carets = []
        table = self.control.table
        for i in range(0,8192,128):
            start = i
            end = i + 120
            rc = table.index_to_row_col(start)
            caret = Caret(*rc)
            caret.anchor_start = caret.anchor_initial_start = rc
            rc = table.index_to_row_col(end)
            caret.anchor_end = caret.anchor_initial_end = rc
            print(caret)
            ch.carets.append(caret)
        flags = self.control.create_mouse_event_flags()
        self.control.commit_change(flags)



#!/usr/bin/env python
"""

"""

import os
import numpy as np
import subprocess
import tempfile

def create_binary(filename, num, outfile, options):
    """Create patterned binary data with the first 7 characters of the filename
    interleaved with a byte ramp, e.g.  A128   \x00A128   \x01A128   \x02 etc.
    """
    root, _ = outfile.split(".")
    prefix = ("%s        " % root)[0:8]
    a = np.fromstring(prefix, dtype=np.uint8)
    b = np.tile(a, (num / np.alen(a)) + 1)[0:num]
    b[7::8] = np.arange(np.alen(b) / 8, dtype=np.uint8)
    with open(filename, "wb") as fh:
        fh.write(b.tostring())

def num_to_letter(num):
    text = []
    while True:
        num, rem = divmod(num, 26)
        text[0:0] = chr(ord("A") + rem)
        if num == 0:
            break
    return "".join(text)

def get_filename(num, used):
    if num not in used:
        used[num] = 0
    prefix = num_to_letter(used[num])
    filename = "%s%d.DAT" % (prefix, num)
    used[num] += 1
    return filename

def franny(filename, *options):
    cmd = ["franny", filename]
    cmd.extend(options)
    msg = subprocess.Popen(cmd, stdout=subprocess.PIPE).communicate()[0]

if __name__ == "__main__":
    import sys
    import argparse
    
    parser = argparse.ArgumentParser(description="Extract images off ATR format disks")
    parser.add_argument("-v", "--verbose", default=0, action="count")
    parser.add_argument("-p", "--prefix", default="B", action="store")
    parser.add_argument("-f", "--filesystem", default="a", action="store")
    parser.add_argument("-d", "--density", default="s", action="store")
    parser.add_argument("-s", "--sectors", default=720, type=int, action="store")
    parser.add_argument("-t", "--template", default="", action="store")
    parser.add_argument("-o", "--output", default="a.out", action="store")
    parser.add_argument("files", metavar="NUM", nargs="+", help="list of numbers specifying the size of each binary file")
    options, extra_args = parser.parse_known_args()
    
    out = options.output
    if options.template:
        franny(out, "-C", "-f", options.filesystem, "-t", options.template)
    else:
        franny(out, "-C", "-f", options.filesystem, "-d", options.density, "-s", str(options.sectors))
    fh, binfile = tempfile.mkstemp()
    os.close(fh)
    used = {}
    files = []
    print(f"Creating image: {out}")
    for entry in options.files:
        entry = entry.lower()
        if entry.startswith("d"):
            index = int(entry[1:])
            outfile = files[index]
            franny(out, "-U", outfile)
        else:
            if "*" in entry:
                repeat, num = entry.split("*", 1)
                num = int(num)
                repeat = int(repeat)
                entries = [num] * repeat
            elif "-" in entry:
                first, last = entry.split("-", 1)
                first = int(first)
                if "," in last:
                    last, step = last.split(",", 1)
                    step = int(step)
                else:
                    step = 1
                last = int(last)
                entries = range(first, last + 1, step)
            else:
                num = int(entry)
                entries = [num]
            for i in entries:
                outfile = get_filename(i, used)
                create_binary(binfile, i, outfile, options)
                franny(out, "-A", "-i", binfile, "-o", outfile)
                files.append(outfile)
    os.unlink(binfile)
